{
  # ---------------------------------------------------------------------- #
  # - Created    : Wednesday, July 25th 2001, 12:15 CET                  - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 12th of October 2003, 19:42 CET                       - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Making the use of various markup and non-markup       - #
  # -              documents as templates for CGI scripts easy.          - #
  # -                                                                    - #
  # - Methods    : getStatus()                                           - #
  # -              getMessage()                                          - #
  # -              getTemplate()                                         - #
  # -              printTemplate()                                       - #
  # -              setValue(variable, value)                             - #
  # -              getValue(variable)                                    - #
  # -              setAttribute(attributename, attributevalue)           - #
  # -              getAttribute(attributename)                           - #
  # ---------------------------------------------------------------------- #
 #
 package Template ;
                                        # 'Constants'
 my($VERSION)  = '2'  ;
 my($RELEASE)  = '0'  ;
 my($PATCH)    = 'b14' ;

 my($REVISION) = "${VERSION}.${RELEASE}.${PATCH}" ;


  # ---------------------------------------------------------------------- #
  # - Created    : Wednesday, July 25th 2001, 12:17 CET                  - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 24th of August 2003, 17:07 CET                        - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Constructor for the Template library.                 - #
  # - Parameters : $type    - the type of the class.                     - #
  # -              $argList - an anonymous hash containing other         - #
  # -                         arguments.                                 - #
  # -                                                                    - #
  # - Returns    : $self    - a reference to an object of this type.     - #
  # ---------------------------------------------------------------------- #
 #
 sub new {
  my($type)    = shift() ;
  my($argList) = shift() ;

  my($self)    = {
                                        # Hash to store "private"
                                        # variables, ie object attributes
                                        # not meant for public consumption.
   'private' => {
                                        # "cooked", as opposed to "raw", is
                                        # the end result after the template
                                        # has been parsed.
                  'cooked' => '',
                                        # "raw" is the complete template
                                        # code as read from file; before any
                                        # processing takes place.
                  'raw'    => '',
                                        # The status structure holds the
                                        # latest status message as well as
                                        # diagnostics output.
                  'status' => {
                               'status'      => 'ok',
                               'diagnostics' => []
                  },
   },
                                        # Hash to store "public" variables,
                                        # ie. object attributes that can be
                                        # manipulated by the setAttribute()
                                        # method, and available to users.
   'public'  => {
                                        # The 'registry' is the structure in
                                        # which all the template references
                                        # are stored.
                  'registry'          => {},
   }
  } ;
  bless($self, $type) ;
                                        # The debug attribute control
                                        # whether or not the _changeStatus()
                                        # method should report what it is
                                        # doing on STDERR.
  $self->setAttribute('debug', 
                       $argList->{'debug'} || 0) ;

                                        # The name of the file in which to
                                        # find the template code. If this
                                        # name does not start with a slash,
                                        # then it will be added to
                                        # 'sourcePath' at a later stage to
                                        # complete the name.
  $self->setAttribute('sourceFile',
                       $argList->{'sourceFile'}) ;

                                        # sourcePath is the name of a
                                        # directory in which to find the
                                        # above sourceFile. Unless
                                        # specified it will default to
                                        # PATH_TRANSLATED.
  $self->setAttribute('sourcePath',
                       $argList->{'sourcePath'} ||
                        $ENV{'PATH_TRANSLATED'}  ||
                         './' ) ;
                                        # expandEmptyItems set to 'no' means
                                        # that as opposed to the default, any
                                        # callback that generates an empty item
                                        # list will still be wrapped in start
                                        # and end tags.
  $self->setAttribute('expandEmptyItems',
                        $argList->{'expandEmptyItems'} || 'no') ;


                                        # If this is set to 'yes', then default
                                        # values will match even if they are
                                        # substrings of actual values.
  $self->setAttribute('matchDefaultSubstring',
                        $argList->{'matchDefaultSubstring'} || 'no') ;

                                        # When the print() method is called,
                                        # the value of this attribute is
                                        # used as the filehandle to send
                                        # the data.
  $self->setAttribute('outputPath',
                       $argList->{'outputPath'} || \*STDOUT) ;

                                        # IF the source file has been
                                        # specified allready, read it.
  if ( $self->getAttribute('sourceFile') ne '' ) {
   if ( ! defined $self->_readSource() ) {
    return($self) ;
   } #end if
  } #end if

  $self->_changeStatus('ok', 'leaving Template.new()') ;
  return($self) ;
 } #end function new()



  # ------------------- #
  # - Private Methods - #
  # ------------------- #

  # ---------------------------------------------------------------------- #
  # - Created    : 8th of March 2003, 22:46 CET                          - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 15th of March 2003, 12:54 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Changes the current status message, and adds one line - #
  # -              of diagnostics output.                                - #
  # -                                                                    - #
  # - Parameters : $self    - a reference to an object of this type.     - #
  # -              $status  - a short status message, default is "ok"    - #
  # -              $message - a diagnostic message.                      - #
  # -                                                                    - #
  # - Returns    : -                                                     - #
  # ---------------------------------------------------------------------- #
 #
 sub _changeStatus {
  my($self)    = shift() ;
  my($status)  = shift() ;
  my($message) = shift() ;

  $self->{'private'}->{'status'}->{'status'} = $status || 'ok' ;
  push(@{$self->{'private'}->{'status'}->{'diagnostics'}},
         scalar(localtime(time)) . " " . $message) ;

  if ( $self->{'public'}->{'debug'} == 1 ) {
   print(STDERR "[debug] [$status] $message\n") ;
  } #end if

 } #end function changeStatus()


  # ---------------------------------------------------------------------- #
  # - Created    : Wednesday, July 25th 2001, 16:03 CET                  - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 20th of March 2003, 10:03 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Read a template source file.                          - #
  # -                                                                    - #
  # - Parameters : Reference: a reference to an object of this type.     - #
  # -                                                                    - #
  # - Returns    : -                                                     - #
  # ---------------------------------------------------------------------- #
 #
 sub _readSource {
  my($self) = shift() ;
  $self->_changeStatus('ok', 'entering Template.readSource()') ;

  my($fileName) = $self->getAttribute('sourceFile') ;
  my($filePath) = $self->getAttribute('sourcePath') ;
  if ( $filePath =~ m#/$# ) {
   $filePath =~ s#/+$## ;
  } #end if
                                        # If no source file has been specified,
                                        # there is absolutely no point in
                                        # proceeding.
  if ( $fileName eq '' ) {
   $self->_changeStatus('error', 'no source file specified') ;
   return(undef()) ;
  } #end if
                                        # If the file cannot be found with
                                        # this name where the script is
                                        # running, add the "filePath"
                                        # component and try again.
  if ( ! -e $fileName ) {
   $fileName = "$filePath/$fileName" ;
   if ( ! -e $fileName ) {
    $self->_changeStatus('error',
                           "source file \"$fileName\" does not exist") ;
    return(undef()) ;
   } #end if
  } #end if
                                        # If the file is empty, do the
                                        # appropriate magic.
  if ( -z $fileName ) {
   $self->_changeStatus('error', 'source file is empty') ;
   return(undef()) ;
  } #end if

                                        # Then read.
  $self->_changeStatus('ok', " opening file \"$fileName\"") ;
  if ( ! open(PTR, "<$fileName") ) {
   $self->_changeStatus('error', " unable to open \"$fileName\" - $!") ;
   return(undef()) ;
  } #end if
                                        # sysread() in everything in the
                                        # file, and close.
  my($buffer) = '' ; my($tmp)    = '' ;
  while( sysread(PTR, $tmp, 250000) ) { 
   $buffer .= $tmp ; $tmp = '' ; 
  } #end while

  $self->_changeStatus('ok', " closing file \"$fileName\"") ;
  close(PTR) ;
                                        # Store the raw data.
  $self->{'private'}->{'raw'}    = $buffer ;

                                        # And make a working copy.
  $self->{'private'}->{'cooked'} = $buffer ;
  $self->_changeStatus('ok', 'leaving Template.readSource()') ;
  return(1) ;
 } #end readSource()


  # ---------------------------------------------------------------------- #
  # - Created    : Wednesday, July 25th 2001, 16:33 CET                  - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 12th of May 2003, 17:57 CET                           - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Boil, that is cook, that is expand the source.        - #
  # -                                                                    - #
  # - Parameters : Reference : a reference to an object of this type.    - #
  # -                                                                    - #
  # - Returns    : -                                                     - #
  # ---------------------------------------------------------------------- #
 #
 sub _boilTemplate {
  my($self) = shift() ;
  $self->_changeStatus('ok', 'in _boilTemplate()') ;

                                        # A local copy of this will be most
                                        # beneficial; as I'd otherwise
                                        # destroy the original data.
  my($templateCode) = $self->{'private'}->{'raw'} ;

                                        # The first step is to collect all
                                        # thetemplate blocks that attempt to
                                        # include other files.
  foreach my $include ( $templateCode =~ /
                                         (
                                          <!--
                                          \s+?
                                          template
                                          \s+?
                                          include
                                          .*?
                                          -->
                                         )
                                        /ixgs ) {
   $self->template_include($include) ;
  } #end foreach
                                        # Then we gather up and store all
                                        # the variables, preferably from
                                        # the cooked version of the template
  foreach my $variable ( $self->{'private'}->{'cooked'} =~ /
                                                           (
                                                            <!--
                                                            \s+?
                                                            template
                                                            \s+?
                                                            set
                                                            .*?
                                                            -->
                                                            )
                                                           /ixgs ) {
   $self->template_set($variable) ;
  } #end foreach


                                        # Here's the really complicated bit.
  foreach my $callback ( $self->{'private'}->{'cooked'} =~ /
                                                         (
                                                          <!--
                                                          \s+?
                                                          template
                                                          \s+?
                                                          callback
                                                          .*?
                                                          -->
                                                         )
                                                        /ixgs ) {
   $self->template_callback($callback) ;
  } #end foreach

                                        # Finally, as the last thing to be
                                        # done, go through the remaining
                                        # variables and expand those that
                                        # are registered.
  foreach my $variable ( $self->{'private'}->{'cooked'} =~ /
                                                            %%
                                                            ([^%]+?)
                                                            %%
                                                           /xsg ) {
   if ( defined($self->getValue($variable)) ) {
    $self->{'private'}->{'cooked'} =~ 
                       s/%%\Q$variable\E%%/$self->getValue($variable)/eg ;
   } #end if
  } #end foreach

  $self->_changeStatus('ok', 'out _boilTemplate()') ;
  return(defined(1)) ;
 } #end function boilTemplate()



  # ---------------------------------------------------------------------- #
  # - Created    : 18th of March 2003, 21:32 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 18th of March 2003, 22:29 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Includes external content at a certain point in the   - #
  # -              template.                                             - #
  # -                                                                    - #
  # - Parameters : String : a reference to an object of this type.       - #
  # -              String : an include construct.                        - #
  # -                                                                    - #
  # - Returns    : -                                                     - #
  # ---------------------------------------------------------------------- #
 #
 sub template_include {
  my($self)    = shift() ;
  my($include) = shift() ;
                                        # The content of 'replacementText'
                                        # will - at a later time - be used
                                        # to replace the original template
                                        # construct.
  my($replacementText) = '' ;
  my($sourceURI) = $include =~ /
                                src
                                \s*
                                =
                                \s*
                                (?:'|")
                                (.*?)
                                (?:'|")
                               /xs ;

  if ( ($sourceURI !~ m#^/#) &&
        ($self->getAttribute('sourcePath') ne '') ) {
   $sourceURI = $self->getAttribute('sourcePath') . "/$sourceURI" ;
  } #end if

  if ( ! open(filePTR, "<$sourceURI") ) {
   $self->_changeStatus('error', "Unable to include $sourceURI: $!") ;
   $replacementText = "**** Unable to include $sourceURI: $! ****" ;
  } else {
   sysread(filePTR, $replacementText, 250000) ;
   close(filePTR) ;
  } #end if-then-else
                                        # Replace the original template
                                        # construct with whatever is in the
                                        # replacementText variable.
  $self->{'private'}->{'cooked'} =~ s/$include/$replacementText/es ;
  return(defined(1)) ;
 } #end function template_include()


  # ---------------------------------------------------------------------- #
  # - Created    : 18th of March 2003, 22:35 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 19th of March 2003, 10:14 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Gathers up variable definitions from the template.    - #
  # -                                                                    - #
  # - Parameters : String : a reference to an object of this type.       - #
  # -              String : a variable definition construct.             - #
  # -                                                                    - #
  # - Returns    : -                                                     - #
  # ---------------------------------------------------------------------- #
 #
 sub template_set {
  my($self)        = shift() ;
  my($variableDef) = shift() ;

  my($varName) = $variableDef =~ /
                                  name
                                  \s*
                                  =
                                  \s*
                                  (?:'|")
                                  (.*?)
                                  (?:'|")
                                 /xs ;

  my($varValue) = $variableDef =~ /
                                  value
                                  \s*
                                  =
                                  \s*
                                  (?:'|")
                                  (.*?)
                                  (?:'|")
                                 /xs ;

                                        # Register the variable.
  $self->register({
                   'name'     => $varName,
                   'callback' => \$varValue
                  }) ;
                                        # And remove the template code that
                                        # did the definition.
  $self->{'private'}->{'cooked'} =~ s/$variableDef//s ;
  return(defined(1)) ;
 } #end function template_set()



  # ---------------------------------------------------------------------- #
  # - Created    : 19th of March 2003, 11:01 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 12th of October 2003, 19:42 CET                       - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Processes callbacks defined in the template.          - #
  # -                                                                    - #
  # - Parameters : String : a reference to an object of this type.       - #
  # -              String : a callback definition construct.             - #
  # -                                                                    - #
  # - Returns    : -                                                     - #
  # ---------------------------------------------------------------------- #
 #
 sub template_callback {
  my($self)        = shift() ;
  my($callbackDef) = shift() ;
                                        # The content of 'replacementText'
                                        # will - at a later time - be used
                                        # to replace the original template
                                        # construct.
  my($replacementText) = '' ;

  my($callbackName) = $callbackDef =~ /
                                       name
                                       \s*
                                       =
                                       \s*
                                       (?:'|")
                                       (.*?)
                                       (?:'|")
                                      /xs ;

  if ( ! defined($self->{'public'}->{'registry'}->{$callbackName}) ) {
   $replacementText =
                    "**** $callbackName is not a registered callback ****" ;
  } #end if
                                        # Parse the start tag construct.
  my($startTag) = '' ; my($startTagQuote) = '' ;
  if ( $callbackDef =~ /starttag/i ) {
   ($startTagQuote) = $callbackDef =~ /starttag\s*=\s*('|")/is ;
   if ( $startTagQuote ne '' ) {
    ($startTag) = $callbackDef =~ /
                                   starttag
                                   \s*
                                   =
                                   \s*
                                   $startTagQuote
                                   ([^$startTagQuote]*)
                                   $startTagQuote
                                  /ixs ;
   } #end if
  } #end if
                                        # Parse the end tag construct
  my($endTag) = '' ; my($endTagQuote) = '' ;
  if ( $callbackDef =~ /endtag/i ) {
   ($endTagQuote) = $callbackDef =~ /endtag\s*=\s*('|")/is ;
   if ( $endTagQuote ne '' ) {
    ($endTag) = $callbackDef =~ /
                                 endtag
                                 \s*
                                 =
                                 \s*
                                 $endTagQuote
                                 ([^$endTagQuote]*)
                                 $endTagQuote
                                /ixs ;
   } #end if
  } #end if
                                        # Parse out item declarations.
  my($itemCode) = '' ; my($itemQuote) = '' ;
  if ( $callbackDef =~ /item/i ) {
   ($itemQuote) = $callbackDef =~ /\s+item\s*=\s*('|")/is ;
   if ( $itemQuote ne '' ) {
    ($itemCode) = $callbackDef =~ /
                                   \s+item
                                   \s*
                                   =
                                   \s*
                                   $itemQuote
                                   ([^$itemQuote]*)
                                   $itemQuote
                                  /ixs ;
   } #end if
  } #end if


                                        # And finally default items.
  my($selectedItemCode) = '' ; my($selectedItemQuote) = '' ;
  if ( $callbackDef =~ /selecteditem/i ) {
   ($selectedItemQuote) = $callbackDef =~ /selecteditem\s*=\s*('|")/is ;
   if ( $selectedItemQuote ne '' ) {
    ($selectedItemCode) = $callbackDef =~ /
                                           selecteditem
                                           \s*
                                           =
                                           \s*
                                           $selectedItemQuote
                                           ([^$selectedItemQuote]*)
                                           $selectedItemQuote
                                          /ixs ;
   } #end if
  } #end if


  my($callbackRef) =
           $self->{'public'}->{'registry'}->{$callbackName}->{'ref'} || '' ;
  my($callbackArg) =
           $self->{'public'}->{'registry'}->{$callbackName}->{'arg'} || undef ;
  my($callbackData) = '' ;

                                        # If the "callback reference" is a
                                        # reference to a piece of code, then
                                        # we'll run that first, and handle it
                                        # later.
  if ( ref($callbackRef) eq 'CODE' ) {
   if ( defined $callbackArg ) {
    $callbackRef = &{$callbackRef}($callbackArg) ;
   } else {
    $callbackRef = &{$callbackRef} ;
   }
  } #end if
                                        # If this is not a reference, I'll
                                        # push the content into a reference to
                                        # anonymous array.
  if ( ref($callbackRef) eq '' ) {
   $callbackData = [$callbackRef] ;
                                        # If ref() returns 'SCALAR' for this
                                        # callback reference, then we'll
                                        # assume a reference to a scalar
                                        # variable, and deref.
  } elsif ( ref($callbackRef) eq 'SCALAR' ) {
   $callbackData = [${callbackRef}] ;
                                        # If this is a reference to an array,
                                        # it needs no massage.
  } elsif ( ref($callbackRef) eq 'ARRAY' ) {
   $callbackData = $callbackRef ;
                                        # If data is a hash refernce pack it
                                        # within a array reference.
  } elsif ( ref($callbackRef) eq 'HASH' ) {
   $callbackData = [$callbackRef] ;
                                        # Unsupported data type.
  } else {
     my $message = "Unsupported data type: ". ref($callbackRef);
     $self->_changeStatus('error', $message) ;
     $self->{'private'}->{'cooked'} = "**** $message ****" ;
     return undef;
  } #end if


                                        # In order to handle default values,
                                        # I'll use ONE reference ...
  my($defaultRef) = 
   $self->{'public'}->{'registry'}->{$callbackName}->{'defaultref'} || '' ;
  my($defaultArg) =
   $self->{'public'}->{'registry'}->{$callbackName}->{'defaultarg'} || undef ;
  my($defaultKey) =
   $self->{'public'}->{'registry'}->{$callbackName}->{'defaultkey'} || '' ;
  my($defaultValue) = '' ;
  
                                        # If the default "callback reference" is
                                        # a reference to a piece of code, then
                                        # we'll run that first, and handle it
                                        # later.
  if ( ref($defaultRef) eq 'CODE' ) {  
   if ( defined $defaultArg ) {
    $defaultRef = &{$defaultRef}($defaultArg) ;
   } else {
    $defaultRef = &{$defaultRef} ;
   }
  } #end if
                                        # If this is not a reference, I'll
                                        # push the content into a reference to
                                        # anonymous array.
  if ( ref($defaultRef) eq '' ) {      
   $defaultValue = [$defaultRef] ;
                                        # If ref() returns 'SCALAR' for this
                                        # callback reference, then we'll
                                        # assume a reference to a scalar
                                        # variable, and deref.
  } elsif ( ref($defaultRef) eq 'SCALAR' ) {
   $defaultValue = [${defaultRef}] ;
                                        # If this is a reference to an array,
                                        # it needs no massage.
  } elsif ( ref($defaultRef) eq 'ARRAY' ) { 
   $defaultValue = $defaultRef ;


  } elsif ( ref($defaultRef) eq 'HASH' ) {
   $defaultValue = [$defaultRef] ;
                                        # Unsupported data type.
  } else {
     my $message = "Unsupported default data type: ". ref($defaultRef);
     $self->_changeStatus('error', $message) ;
     $self->{'private'}->{'cooked'} = "**** $message ****" ;
     return undef;
  } #end if
                                        # If the data is a plain list
                                        # (array), we create one copy of
                                        # the itemCode per list item, and
                                        # replaced from left to right.
  if ( ref($callbackData) eq 'ARRAY' ) {

                                        # De-reference it. At this point in
                                        # time the data is in a normal array
                                        # called 'content'.
   my(@content) = @{$callbackData} ;
                                        # On occation it is desired to let a
                                        # default value match only the exact
                                        # real value, and not a substring.
   my $matchSubstring = $self->getAttribute('matchDefaultSubstring');

                                        # Traverse the list of data items, and
                                        # produce one item of code per.
   for ( my $i = 0 ; $i <= $#content ; $i++ ) {
    my($entry)   = $content[$i] ;
    my($tmpCode) = $itemCode ;
                                        # If the entry is not a reference, then
                                        # it is likely to be a plain scalar,
                                        # and we replace the first variable
                                        # in the item code with it's value.
    if ( ref($entry) eq '' ) {
     my $tmpdefValue = $defaultValue->[0];
     if ( $matchSubstring eq 'yes' ) {
                                        # Do a substring match.
      if ( "$entry" =~ /\Q$tmpdefValue\E/ ) {
       $tmpCode = $selectedItemCode ;
      } #end if
     } else {
                                        # Do a exact string match.
      if ( "$entry" eq "$tmpdefValue" ) {
       $tmpCode = $selectedItemCode ;
      } #end if
     } #end if
     $tmpCode =~ s/%%[^%]*%%/$entry/e ;
                                        # If the entry is a reference to an
                                        # array, then I'll make a copy of that
                                        # list, and replace the variables from
                                        # left to right.

    } elsif ( ref($entry) eq 'ARRAY' ) {

     my(@list) = @{$entry} ;
                                        # Check to see if we have the default
                                        # array.
     my @defaultList = @{$defaultValue};
     my $matchCount  = 0;
     for (my $i = 0; $i < @defaultList; $i++ ) {
      my $tmpdefValue = $defaultList[$i];
                                        # Do a sub string match?
      if ( $matchSubstring eq 'yes' ) {
                                        # if the entry slice doesn't match
                                        # its corresponding default value
                                        # we don't have to search anymore in
                                        # in the default value array.
       unless( "$entry->[$i]" =~ /\Q$tmpdefValue\E/ ) {
        last;
       } #end unless
                                        # Do an exact string math.
      } else {
                                        # if the entry slice doesn't match
                                        # its corresponding default value
                                        # we don't have to search anymore in
                                        # in the default value array.
       unless( "$entry->[$i]" eq "$tmpdefValue" ) {
        last;
       } #end unless
      } #end if
                                        # The value matched agains the default
                                        # value. So update the match counter.
      $matchCount++;
     } #end for
                                        # If all default value elements
                                        # matched their corrsponding values we
                                        # should use the markup code for the
                                        # "selected item".
     if ( $matchCount == @defaultList ) {
      $tmpCode = $selectedItemCode ;
     } #end if
                                        # Fill in the vaules in the resulting
                                        # markup template code.
     for (my $i = 0 ; $i <= $#list ; $i++ ) {
      $tmpCode =~ s/%%[^%]*%%/$list[$i]/e ;
     } #end for
                                        # If the entry is a reference to a
                                        # hash, I'll pick out any and all
                                        # variables in the item code, and
                                        # insert the matching value from the
                                        # hash.
    } elsif ( ref($entry) eq 'HASH' ) {
                                        # Assume we have a match and later
                                        # prove this assumption is wrong.
     my $Match = 1;
                                        # No default key defined. We have
                                        # to match the whole default hash with 
                                        # the entry hash.
     if ( $defaultKey eq '' ) {
      my %defaultHash = %{$defaultValue->[0]};
      while ( my($key, $value) = each %defaultHash ) {
                                        # The key doesn't exist in the entry
                                        # entry data. Then there is no point
                                        # in proceding.
       unless( exists $entry->{$key} ) {
        $Match = undef;
        last;
       } #end unless
                                        # Do a sub string match?
       if ( $matchSubstring eq 'yes' ) {

        unless ( "$entry->{$key}" =~ /\Q$value\E/ ) {
         $Match = undef;
         last;
        } #end unless
                                        # Do an exact match.
       } else {
        unless ( "$entry->{$key}" eq "$value" ) {
         $Match = undef;
         last;
        } #end unless
       } #end if
      } #end while
                                        # Is the default hash empty?
      unless( %defaultHash ) {
                                        # Then we cannot have a match!
        $Match = undef;
      }

                                        # We have a default key.
     } else {
                                        # The default key is present in
                                        # the current entry.
      if ( exists $entry->{$defaultKey} ) {
       my $value = $defaultValue->[0];
                                        # Do a sub string match?
       if ( $matchSubstring eq 'yes' ) {
       
        unless ( "$entry->{$defaultKey}" =~ /\Q$value\E/ ) {
         $Match = undef;
        } #end unless
                                        # Do an exact match.
       } else {
        unless ( "$entry->{$defaultKey}" eq "$value" ) {
         $Match = undef;
        } #end unless
       } #end if
      } #ned if
     } #end if
                                        # We have a match.
     if ( $Match ) {
      $tmpCode = $selectedItemCode ;
     } #end if

     foreach my $key ( $tmpCode =~ /%%([^%]+?)%%/sg ) {
      $tmpCode =~ s/%%$key%%/$entry->{$key}/eg ;
     } #end foreach
                                        # Unsupported content type.
    } else {
print("xx\n") ;
     my $message = "Unsupported content type: ". ref($entry);
     $self->_changeStatus('error', $message) ;
     $self->{'private'}->{'cooked'} = "**** $message ****" ;
     return undef;
    } #end if
                                        # Finally add the temporary item code
                                        # to the end result.
    $replacementText .= "$tmpCode\n"  ;
   } #end for
  } #end if
                                        # Replace the original template
                                        # construct with whatever is in the
                                        # replacementText variable.
  if ( ($replacementText ne '') ||
        ($self->getAttribute('expandEmptyItems') eq 'yes') ) {
   $replacementText = "$startTag\n $replacementText \n$endTag" ;
  } #end if


  $self->{'private'}->{'cooked'} =~ s/\Q$callbackDef\E/$replacementText/es ;
  return(defined(1)) ;
 } #end function template_callback()





  # ------------------ #
  # - Public Methods - #
  # ------------------ #

  # ---------------------------------------------------------------------- #
  # - Created    : 9th of March 2003, 10:11 CET                          - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 15th of March 2003, 12:45 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Retrieves the current status message.                 - #
  # -                                                                    - #
  # - Parameters : $self    - a reference to an object of this type.     - #
  # -                                                                    - #
  # - Returns    : String: the current status message.                   - #
  # ---------------------------------------------------------------------- #
 #
 sub getStatus {
  my($self) = shift() ;
  return($self->{'private'}->{'status'}->{'status'}) ;
 } #end function getStatus()


  # ---------------------------------------------------------------------- #
  # - Created    : 9th of March 2003, 10:20 CET                          - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 15th of March 2003, 12:46 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Retrieves the current diagnostics message.            - #
  # -                                                                    - #
  # - Parameters : Reference: reference to an object of this type.       - #
  # -                                                                    - #
  # - Returns    : String   : the current diagnostics message.           - #
  # ---------------------------------------------------------------------- #
 #
 sub getMessage {
  my($self)  = shift() ;
  my($lastEntry) = $#{$self->{'private'}->{'status'}->{'diagnostics'}} ;
  return($self->{'private'}->{'status'}->{'diagnostics'}->[$lastEntry]) ;
 } #end function getMessage()


  # ---------------------------------------------------------------------- #
  # - Created    : 9th of March 2003, 10:20 CET                          - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 15th of March 2003, 12:33 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Retrieves - after boiling - the template.             - #
  # -                                                                    - #
  # - Parameters : Reference: reference to an object of this type.       - #
  # -                                                                    - #
  # - Returns    : String   : the cooked, or expanded, template.         - #
  # ---------------------------------------------------------------------- #
 #
 sub getTemplate {
  my($self)  = shift() ;
                                        # First we boil, then we serve.
  $self->_boilTemplate() ;
  return($self->{'private'}->{'cooked'}) ;
 } #end function getTemplate()


  # ---------------------------------------------------------------------- #
  # - Created    : Wednesday, July 25th 2001, 16:36 CET                  - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 9th of March 2003, 11:09 CET                          - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Output the template code.                             - #
  # -                                                                    - #
  # - Parameters : Reference: a reference to an object of this type.     - #
  # -                                                                    - #
  # - Returns    : -                                                     - #
  # -                                                                    - #
  # ---------------------------------------------------------------------- #
  #
  sub printTemplate {
   my($self)  = shift() ;
   $self->_changeStatus('ok', 'entering Template.print()') ;

                                        # Printing is done by first using
                                        # the getTemplate() method to
                                        # retrieve the data, then print it.
   my($fileHandle) = $self->getAttribute('outputPath') ;
   print($fileHandle $self->getTemplate()) ;

   $self->_changeStatus('ok', 'leaving Template.print()') ;
  } #end function printTemplate()


  # ---------------------------------------------------------------------- #
  # - Created    : Wednesday, July 25th 2001, 12:19 CET                  - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 20th of March 2003, 11:12 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Retrieve the value of a pre-defined variable.         - #
  # -                                                                    - #
  # - Parameters : Reference : a reference to an object of this type.    - #
  # -              String    : the name of the variable.                 - #
  # -                                                                    - #
  # - Returns    : String    : the value of the variable, or undef()     - #
  # -                                                                    - #
  # ---------------------------------------------------------------------- #
 #
 sub getValue {
  my($self)         = shift() ;
  my($variableName) = shift() ;
  my($value) = $self->{'public'}->{'registry'}->{$variableName}->{'ref'} ;

  if ( ref($value) eq 'SCALAR' ) {
   return(${$value}) ;
  } #end if

  if ( ! ref($value) ) {
   return($value) ;
  } #end if

  if ( ! defined($value) ) {
   return("") ;
  } #end if

  return("Variable $variableName is not a scalar value!") ;
 } #end getValue()


  # ---------------------------------------------------------------------- #
  # - Created    : Wednesday, July 25th 2001, 12:19 CET                  - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 9th of March 2003, 11:12 CET                          - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Set - rarely - an attribute in the template object.   - #
  # -                                                                    - #
  # - Parameters : Reference : a reference to an object of this type.    - #
  # -              String    : attribute to be set.                      - #
  # -              String    : the value to set.                         - #
  # -                                                                    - #
  # - Returns    : -                                                     - #
  # -                                                                    - #
  # ---------------------------------------------------------------------- #
 #
 sub setValue {
  my($self)      = shift() ;
  my($attribute) = shift() ;
  my($value)     = shift() ;

 } #end setValue()







  # ---------------------------------------------------------------------- #
  # - Created    : 15th of March 2003, 12:12 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 15th of March 2003, 13:06 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Retrieve the value of an object attribute, most often - #
  # -              an argument to new()                                  - #
  # -                                                                    - #
  # - Parameters : Reference : a reference to an object of this type.    - #
  # -              String    : the name of the attribute.                - #
  # -                                                                    - #
  # - Returns    : String    : the value of the attribute, or undef()    - #
  # -                                                                    - #
  # ---------------------------------------------------------------------- #
 #
 sub getAttribute {
  my($self)    = shift() ;
  my($attributeName) = shift() ;
  $self->_changeStatus('ok',
                        'in Template.getValue("' . $attributeName . '")');

  if ( $self->{'public'}->{$attributeName} ne '' ) {
   return($self->{'public'}->{$attributeName}) ;
  } else {
   return(undef()) ;
  } #end if-then-else

  $self->_changeStatus('ok', 'out Template.getValue()') ;
 } #end function getAttribute()



  # ---------------------------------------------------------------------- #
  # - Created    : 15th of March 2003, 12:19 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 20th of March 2003, 09:56 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Set the value of an object attribute.                 - #
  # -                                                                    - #
  # - Parameters : Reference : a reference to an object of this type.    - #
  # -              String    : the name of the attribute.                - #
  # -              String    : the value of the attribute.               - #
  # -                                                                    - #
  # - Returns    : String    : the value of the attribute, or undef()    - #
  # -                                                                    - #
  # ---------------------------------------------------------------------- #
 #
 sub setAttribute {
  my($self)           = shift() ;
  my($attributeName)  = shift() ;
  my($attributeValue) = shift() ;
  $self->{'public'}->{$attributeName} = $attributeValue ;
 } #end function setAttribute()

 

  # ---------------------------------------------------------------------- #
  # - Created    : 9th of March 2003, 11:58 CET                          - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 21st of April 2003, 13:15 CET                         - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Register callback names for later use in a template.  - #
  # -                                                                    - #
  # - Parameters : Reference : a reference to an object of this type.    - #
  # -              Reference : a reference to a hash of argument/value   - #
  # -                          pairs.                                    - #
  # -                                                                    - #
  # - Returns    : -                                                     - #
  #
  # Usage:
  #
  # $Template->register(
  #     'name'         => $Name,
  #     'callback'     => \&Callback, or $Data or \@Data or \%Data
  #     'args'         => \@Args,
  #     'default'      => \&Default, or $Default or \@Default or \%Default
  #     'default_args' => \@Default_args
  # );
  #
  # Where:
  #
  # $Name          template name.
  # &Callback      callback routine.
  # $Data          callback data
  # @Args          optional callback arguments.
  # &Default       default value callback.
  # $Default       default value.
  # @Default_args  optional default value callback arguments.
  # -                                                                    - #
  # ---------------------------------------------------------------------- #
 #
 sub register {
  my($self)   = shift() ;
  my($argRef) = shift() ;
                                        # Fetch the name, and create an empty
                                        # structure for it.
  my($name) = $argRef->{'name'} ;
  $self->{'public'}->{'registry'}->{$name} = {} ;

                                        # Store the normal and default data
                                        # structures.
  $self->{'public'}->{'registry'}->{$name}->{'ref'} =
   $argRef->{'callback'} ;

  $self->{'public'}->{'registry'}->{$name}->{'defaultref'} =
   $argRef->{'default'} ;
                                        # And the arguments to both.
  $self->{'public'}->{'registry'}->{$name}->{'arg'} =
   $argRef->{'callbackArguments'} ;

  $self->{'public'}->{'registry'}->{$name}->{'defaultarg'} =
   $argRef->{'defaultArguments'} ;

                                        # Finally the default key.
  $self->{'public'}->{'registry'}->{$name}->{'defaultkey'} =
   $argRef->{'defaultKey'} ;

  return(defined(1)) ;
 } #end function register()




  # ---------------------------------------------------------------------- #
  # - Created    : 15th of September 2003, 03:11 CET                     - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Updated    : 15th of September 2003, 03:11 CET                     - #
  # -         by : Tina Holmboe <tina@greytower.net>                     - #
  # -                                                                    - #
  # - Purpose    : Reset the cooked version                              - #
  # -                                                                    - #
  # - Parameters : Reference : a reference to an object of this type.    - #
  # -              Reference : a reference to a hash of argument/value   - #
  # -                          pairs.                                    - #
  # -                                                                    - #
  # - Returns    : -                                                     - #
  #
  # ---------------------------------------------------------------------- #
 #
 sub reset {
  my($self)   = shift() ;
  my($argRef) = shift() ;
  $self->_changeStatus('ok', 'in reset()') ;
  $self->{'private'}->{'cooked'} = $self->{'private'}->{'raw'} ;
 } #end method reset()

} #end package Template

1 ;

